#version 330 core

#define MAX_POINT_LIGHTS 100

uniform vec3 cameraPosition;

in vec3 FragPos;
in vec3 Normal;
in vec2 TexCoords;

// Textures
uniform sampler2D albedoMap;
uniform sampler2D normalMap;
uniform sampler2D metallicMap;
uniform sampler2D roughnessMap;
uniform sampler2D aoMap;

// Directional light
uniform struct DirectionalLight
{
	vec3 direction;
	vec3 color;
};
uniform DirectionalLight directionalLight;

// Point light
uniform int pointLightCount;
uniform vec3 pointLightPositions[MAX_POINT_LIGHTS];	
uniform vec3 pointLightColors[MAX_POINT_LIGHTS];
uniform vec3 pointLightAttenuations[MAX_POINT_LIGHTS];

out vec4 FragColor;

const float PI = 3.14159265359;
const float DIRECT_LIGHT_INTENSITY = 10.0;

// Calculate the normal from the normal map by converting from tangent to view space
vec3 getNormalFromMap()
{
    // Read normal from normal map
    vec3 tangentNormal = texture(normalMap, TexCoords).xyz * 2.0 - 1.0;

    vec3 Q1 = dFdx(FragPos);
    vec3 Q2 = dFdy(FragPos);
    vec2 st1 = dFdx(TexCoords);
    vec2 st2 = dFdy(TexCoords);

    vec3 N = normalize(Normal);
    vec3 T = normalize(Q1*st2.t - Q2*st1.t);
    vec3 B = -normalize(cross(N, T));
    mat3 TBN = mat3(T, B, N);

    return normalize(TBN * tangentNormal);
}

// Calculate the approximate distribution of aligned microfacets
float DistributionGGX(vec3 N, vec3 H, float roughness)
{
    float a = roughness*roughness;
    float a2 = a*a;
    float NdotH = max(dot(N, H), 0.0);
    float NdotH2 = NdotH*NdotH;

    float nom = a2;
    float denom = (NdotH2 * (a2 - 1.0) + 1.0);
    denom = PI * denom * denom;

    return nom / denom;
}

// Calculate the geometric attenuation factor, based on selfshadowing of microfacets
float GeometrySchlickGGX(float NdotV, float roughness)
{
    float r = (roughness + 1.0);
    float k = (r*r) / 8.0;

    float nom = NdotV;
    float denom = NdotV * (1.0 - k) + k;

    return nom / denom;
}

// Calculate the geometric attenuation factor, taking both view and light direction into account
float GeometrySmith(vec3 N, vec3 V, vec3 L, float roughness)
{
    float NdotV = max(dot(N, V), 0.0);
    float NdotL = max(dot(N, L), 0.0);
    float ggx2 = GeometrySchlickGGX(NdotV, roughness);
    float ggx1 = GeometrySchlickGGX(NdotL, roughness);

    return ggx1 * ggx2;
}

// Calculate the Fresnel equation
vec3 fresnelSchlick(float cosTheta, vec3 F0)
{
    return F0 + (1.0 - F0) * pow(clamp(1.0 - cosTheta, 0.0, 1.0), 5.0);
}

void main() {
    // Read texture values
    vec3 albedo = pow(texture(albedoMap, TexCoords).rgb, vec3(2.2)); // convert from sRGB to linear
    float metallic = texture(metallicMap, TexCoords).r;
    float roughness = texture(roughnessMap, TexCoords).r;
    float ao = texture(aoMap, TexCoords).r;

    // Determine normal and view direction
    vec3 N = getNormalFromMap();
    vec3 V = normalize(cameraPosition - FragPos);

    // Calculate the reflectance at normal incidence, interpolating between dielectric and metallic based on metallic value  
    vec3 F0 = vec3(0.04); 
    F0 = mix(F0, albedo, metallic);

    // Main reflectance equation (directional light is calculated in the i == pointLightCount case)
    vec3 Lo = vec3(0.0);
    for (int i = 0; i < pointLightCount + 1; i++) {
        // Calculate light direction and half vector
        vec3 L;
        if (i == pointLightCount) {
            L = normalize(-directionalLight.direction);
        } else {
            L = normalize(pointLightPositions[i] - FragPos);
        }
        vec3 H = normalize(V + L);

        // Calculate light radiance
        vec3 radiance;
        if (i == pointLightCount) {
            radiance = directionalLight.color * DIRECT_LIGHT_INTENSITY;
        } else {
            float distance = length(pointLightPositions[i] - FragPos);
            float attenuation = 1.0 / (pointLightAttenuations[i].x + pointLightAttenuations[i].y * distance + pointLightAttenuations[i].z * distance * distance);
            radiance = pointLightColors[i] * attenuation;
        }

        // Cook-Torrance BRDF
        float NDF = DistributionGGX(N, H, roughness);
        float G = GeometrySmith(N, V, L, roughness);
        vec3 F = fresnelSchlick(max(dot(H, V), 0.0), F0);
           
        vec3 numerator = NDF * G * F; 
        float denominator = 4.0 * max(dot(N, V), 0.0) * max(dot(N, L), 0.0) + 0.0001; // + 0.0001 to prevent divide by zero
        vec3 specular = numerator / denominator;
        
        // Specular component is equal to Fresnel
        vec3 kS = F;
        // Get diffuse component based on energy conservation
        vec3 kD = vec3(1.0) - kS;
        // Adjust diffuse component based on metallic value
        kD *= 1.0 - metallic;  

        // Calculate the final outgoing radiance
        float NdotL = max(dot(N, L), 0.0);
        Lo += (kD * albedo / PI + specular) * radiance * NdotL; 
    }
    
    // Basic ambient lighting approximation
    vec3 ambient = vec3(0.1) * albedo * ao;
    
    // Final color
    vec3 color = ambient + Lo;

    // HDR tonemapping
    color = color / (color + vec3(1.0));

    // Gamma correction
    color = pow(color, vec3(1.0 / 2.2)); 

    FragColor = vec4(color, 1.0);
}